unit uSkinFile;

{$MODE Delphi}

interface

uses
  Classes, SysUtils;

type
  PFileInfo = ^TFileInfo;
  TFileInfo = record
    Name          : String;
    Start, Length : Cardinal;
  end;

type
  TSkinFileOnProgress = procedure(const Max, Current: Cardinal) of Object;

  TSkinFile = class
  private
    fFile       : File;
    fFileList   : TList;
    fFilePath   : String;
    fOnProgress : TSkinFileOnProgress;
    procedure Init;
    procedure NewFile(const aName : String; const aStart, aLength : Cardinal);
  public
    constructor Create(const FileName : String);
    destructor Destroy; override;
    procedure Files(FileList : TStrings);
    procedure AddFile(const FileName : String);
    procedure ExtractFile(const FileIndex : Integer; DestFileName : String = ''); overload;
    procedure ExtractFile(const ArchiveFileName : String; DestFileName : String = ''); overload;
    procedure ExtractAll(ExtractPath : String = '');
    property OnProgress : TSkinFileOnProgress read fOnProgress write fOnProgress;
  end;

implementation

constructor TSkinFile.Create(const FileName : String);
begin
  inherited Create;
  fFilePath := ExtractFilePath(FileName);
  if FileExists(FileName) then
  begin
    AssignFile(fFile, FileName);
    Reset(fFile, 1);
  end else
  begin
    AssignFile(fFile, FileName);
    Rewrite(fFile, 1);
  end;
  fFileList := TList.Create;
  Init;
end;

destructor TSkinFile.Destroy;
var i : Integer;
begin
  CloseFile(fFile);
  for i := 0 to fFileList.Count-1 do
    Dispose(PFileInfo(fFileList.Items[i]));
  fFileList.Free;
  inherited;
end;

procedure TSkinFile.Init;
var FN           : String;
    NameLen      : Byte;
    FileLen, Pos : Cardinal;
begin
  Seek(fFile, 0);
  Pos := 0;
  while not EOf(fFile) do
  begin
    BlockRead(fFile, NameLen, SizeOf(NameLen));
    SetLength(FN, NameLen);
    BlockRead(fFile, FN[1], NameLen);
    BlockRead(fFile, FileLen, SizeOf(FileLen));
    inc(Pos, SizeOf(NameLen) + NameLen + SizeOf(FileLen));
    NewFile(FN, Pos, FileLen);
    inc(Pos, FileLen);
    Seek(fFile, Pos);
  end;
end;

procedure TSkinFile.NewFile(const aName : String; const aStart, aLength : Cardinal);
var FileInfo : PFileInfo;
begin
  New(FileInfo);
  with FileInfo^ do
  begin
    Name   := aName;
    Start  := aStart;
    Length := aLength;
  end;
  fFileList.Add(FileInfo);
end;

procedure TSkinFile.Files(FileList : TStrings);
var i : Integer;
begin
  FileList.BeginUpdate;
  for i := 0 to fFileList.Count-1 do
    FileList.Add(PFileInfo(fFileList.Items[i]).Name);
  FileList.EndUpdate;
end;

procedure TSkinFile.AddFile(const FileName : String);
const BufferSize = 1024;
var FN           : String;
    NameLen      : Byte;
    FileLen, Pos : Cardinal;
    Len          : Integer;
    SFile        : File;
    Buffer       : Array[0..BufferSize] of Byte;
begin
  Seek(fFile, FileSize(fFile));

  FN      := ExtractFileName(FileName);
  NameLen := Length(FN);
  BlockWrite(fFile, NameLen, SizeOf(NameLen));
  BlockWrite(fFile, FN[1]  , NameLen);

  AssignFile(SFile, FileName);
  try
    Reset(SFile, 1);
    FileLen := FileSize(SFile);
    BlockWrite(fFile, FileLen, SizeOf(FileLen));

    NewFile(FN, FileSize(fFile), FileLen);

    Pos := 0;
    while not EOf(SFile) do
    begin
      BlockRead(SFile, Buffer, BufferSize, Len);
      BlockWrite(fFile, Buffer, Len);
      inc(Pos, Len);
      if Assigned(fOnProgress) then
        fOnProgress(FileLen, Pos);
    end;
  finally
    CloseFile(SFile);
  end;
end;

procedure TSkinFile.ExtractFile(const FileIndex : Integer; DestFileName : String = '');
const BufferSize = 1024;
var FileInfo : PFileInfo;
    DFile    : File;
    Pos      : Cardinal;
    Len      : Integer;
    Buffer   : Array[0..BufferSize] of Byte;
begin
  if FileIndex < fFileList.Count then
    FileInfo := PFileInfo(fFileList.Items[FileIndex])
  else
    Exit;
  if DestFileName = '' then
    DestFileName := fFilePath + FileInfo.Name;

  Seek(fFile, FileInfo.Start);

  Assign(DFile, DestFileName);
  try
    Rewrite(DFile, 1);
    Pos := 0;
    while Pos < FileInfo.Length do
    begin
      if FileInfo.Length-Pos > BufferSize then
        Len := BufferSize
      else
        Len := FileInfo.Length-Pos;
      BlockRead(fFile, Buffer, Len);
      BlockWrite(DFile, Buffer, Len);
      inc(Pos, Len);
      if Assigned(fOnProgress) then
        fOnProgress(FileInfo.Length, Pos);
    end;
  finally
    CloseFile(DFile);
  end;
end;

procedure TSkinFile.ExtractFile(const ArchiveFileName : String; DestFileName : String = '');
var i : Integer;
begin
  for i := 0 to fFileList.Count-1 do
    if PFileInfo(fFileList.Items[i]).Name = ArchiveFileName then
      ExtractFile(i, DestFileName);
end;

procedure TSkinFile.ExtractAll(ExtractPath: String = '');
var i : Integer;
begin
  if (ExtractPath <> '') and (ExtractPath[Length(ExtractPath)] <> '\') then
    ExtractPath := ExtractPath + '\';
  for i := 0 to fFileList.Count-1 do
    if ExtractPath = '' then
      ExtractFile(i)
    else
      ExtractFile(i, ExtractPath + PFileInfo(fFileList.Items[i]).Name);
end;

end.
